#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdexcept>  
#include <iostream>   
#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>  
#endif


#include <eo>
#include <ga.h>


#include "fitness.h"


using namespace std;




void main_function(int argc, char **argv)
{

// PARAMETRES

   unsigned int SEED;      // seed for random number generator
  //  const unsigned int T_SIZE = 3;     // size for tournament selection
   unsigned int POP_SIZE;  // Size of population
   unsigned int MAX_GEN;  // Maximum number of generation before STOP
   float CROSS_RATE;      // Crossover rate
   double P_MUT_PER_BIT; // probability of bit-flip mutation
   float MUT_RATE;        // mutation rate
   double EPSILON = 1.0;  // range for real uniform mutation

// PARAMETRES
    eoParser parser(argc, argv);

    // For each parameter, define Parameter, read it through the parser,
    // and assign the value to the variable
    eoValueParam<uint32_t> seedParam(time(0), "seed", "Random number seed", 'S');
    parser.processParam( seedParam );
    SEED = seedParam.value();


   // parameters for evolution engine
    eoValueParam<unsigned int> popSizeParam(50, "popSize", "Population size",'P');
    parser.processParam( popSizeParam, "Evolution engine" );
    POP_SIZE = popSizeParam.value();

    // init and stop
    eoValueParam<unsigned int> maxGenParam(1000, "maxGen", "Maximum number of generations",'G');
    parser.processParam( maxGenParam, "Stopping criterion" );
    MAX_GEN = maxGenParam.value();

   // operators probabilities at the algorithm level
    eoValueParam<double> pCrossParam(0.6, "pCross", "Probability of Crossover", 'C'); 
    parser.processParam( pCrossParam, "Genetic Operators" );
    CROSS_RATE = pCrossParam.value();

    eoValueParam<double> pMutParam(0.2, "pMut", "Probability of Mutation", 'M');
    parser.processParam( pMutParam, "Genetic Operators" );
    MUT_RATE = pMutParam.value();

   // relative rates and private parameters for mutations;
    eoValueParam<double> pMutPerBitParam(0.01, "pMutPerBit", "Probability of flipping 1 bit in bit-flip mutation", 'b');
    parser.processParam( pMutPerBitParam, "Genetic Operators" );
    P_MUT_PER_BIT = pMutPerBitParam.value();

    // name of file to save statistics 
    eoValueParam<string> saveStatsParam("fitness.log", "LogFile","A file to log statistics",'L');
    parser.processParam( saveStatsParam, "Log Statistics File" );
    string saveStatsName = saveStatsParam.value();


    // the name of the "status" file where all actual parameter values will be saved
    string str_status = parser.ProgramName() + ".status"; // default value
    eoValueParam<string> statusParam(str_status.c_str(), "status","Status file",'S');
    parser.processParam( statusParam, "Persistence" );

   // do the following AFTER ALL PARAMETERS HAVE BEEN PROCESSED
   // i.e. in case you need parameters somewhere else, postpone these
    if (parser.userNeedsHelp())
      {
        parser.printHelp(cout);
        exit(1);
      }
    if (statusParam.value() != "")
      {
	ofstream os(statusParam.value().c_str());
	os << parser;		// and you can use that file as parameter file
      }


  // GENERAL
  //////////////////////////
  //  Random seed
  //////////////////////////
  //reproducible random seed: if you don't change SEED above, 
  // you'll aways get the same result, NOT a random run
  rng.reseed(SEED);

  // EVAL
  /////////////////////////////
  // Fitness function
  ////////////////////////////
  // Evaluation: from a plain C++ fn to an EvalFunc Object
  eoEvalFuncPtr<Indi> eval(  fitness_value );


   // INIT
  ////////////////////////////////
  // Initilisation of population
  ////////////////////////////////

  // declare the population
  eoPop<Indi> pop;
  // fill it!
  for (unsigned int igeno=0; igeno<POP_SIZE; igeno++)
    {
      Indi v;           // void individual, to be filled
      for (unsigned ivar=0; ivar<2; ivar++)
        {
	  double r = 10*rng.uniform()-5;
          v.push_back(r);      // append that random value to v
        }
      eval(v);                 // evaluate it
      pop.push_back(v);        // and put it in the population
    }

  // OUTPUT
  // sort pop before printing it!
  //  pop.sort();
  // Print (sorted) intial population (raw printout)
  //cout << "Initial Population" << endl;
  //cout << pop;

  // ENGINE
  /////////////////////////////////////
  // selection and replacement
  ////////////////////////////////////
  // SELECT
  // The roulette wheel selection
  eoProportionalSelect<Indi> selectOne;  
  eoSelectPerc<Indi> select( selectOne ); 

  double elita_percentage = 0.05;                                                                 
  eoDeterministicSaDReplacement<Indi> replace(elita_percentage,1.0-elita_percentage);
                                                          
 

  // OPERATORS
  //////////////////////////////////////
  // The variation operators
  //////////////////////////////////////
  // CROSSOVER
  // offspring(i) is a linear combination of parent(i)
  eoSegmentCrossover<Indi> xover;

  // MUTATION
  // offspring(i) uniformly chosen in [parent(i)-epsilon, parent(i)+epsilon]
  eoUniformMutation<Indi>  mutation(EPSILON); 

 // The operators are  encapsulated into an eoTRansform object
  eoSGATransform<Indi> transform(xover, CROSS_RATE, mutation, MUT_RATE);


  // STOP
  // CHECKPOINT
  //////////////////////////////////////
  // termination condition
  /////////////////////////////////////
  // stop after MAX_GEN generations

   eoGenContinue<Indi> genCont(MAX_GEN);

 // but now you want to make many different things every generation 
 // (e.g. statistics, plots, ...).
 // the class eoCheckPoint is dedicated to just that:
 // Declare a checkpoint (from a continuator: an eoCheckPoint 
 // IS AN eoContinue and will be called in the loop of all algorithms)
 eoCheckPoint<Indi> checkpoint(genCont);

  // Create a counter parameter
     eoValueParam<unsigned> generationCounter(0, "Gen.");

   // Create an incrementor (sub-class of eoUpdater). Note that the 
     // parameter's value is passed by reference, 
     // so every time the incrementer is updated (every generation),
     // the data in generationCounter will change.
     eoIncrementor<unsigned> increment(generationCounter.value());
  // Add it to the checkpoint, 
     // so the counter is updated (here, incremented) every generation
     checkpoint.add(increment);
  // now some statistics on the population:
     // Best fitness in population
     eoBestFitnessStat<Indi> bestStat;
     // Second moment stats: average and stdev
     eoSecondMomentStats<Indi> SecondStat;
   // Add them to the checkpoint to get them called at the appropriate time
     checkpoint.add(bestStat);
     checkpoint.add(SecondStat);
    
     // A file monitor: will print parameters to ... a File, yes, you got it!
     eoFileMonitor fileMonitor(saveStatsName, " ");

     // the checkpoint mechanism can handle multiple monitors
     checkpoint.add(fileMonitor);
     // the fileMonitor can monitor parameters, too, but you must tell it!
     fileMonitor.add(generationCounter);
     fileMonitor.add(bestStat);
     fileMonitor.add(SecondStat);

   
   
  // GENERATION
  /////////////////////////////////////////
  // the algorithm
  ////////////////////////////////////////
     eoEasyEA<Indi> gga(checkpoint,
			eval,
			select,
			transform,
			replace);

  // Apply algo to pop - that's it!
     cout << "START" << endl;
     gga(pop);
     cout << "END" << endl;
     cout << pop.best_element() << endl;
  
  // OUTPUT
  // Print (sorted) intial population
  //  pop.sort();
  //cout << "FINAL Population\n" << pop << endl;
  // GENERAL
}
 


int main(int argc, char **argv)
{

    try
    {
        main_function(argc, argv);
    }
    catch(exception& e)
    {
        cout << "Exception: " << e.what() << '\n';
    }

    return 1;
}
